/*
 * Sistemas de Tempo Real, 2009/2010
 * Paulo Pedreiras, Luis Almeida
 * Sept 2009
 *
 * STR Linux/RTAI lab demo
 * 	4 RT tasks toggle a bit on the paralell port.
 * 	Bit status is sent to rt sharem memory
 * 	Linux user-space task reads rt shared memory and displays task status
 *
 * 	RT module
 *
 * Changelog:
 *  2006/09/21, Paulo Pedreiras
 *     	Ported from RTAI 1.0 to RTAI 3.3
 *     	Major changes: thread arg in init changed from int to long
 *     	task_preempt_always() not supported anymore
 *     	Makefile updated due to change in linux kernel from 2.4 to 2.6
 *     	macro to define init/cleanup functions (kernel 2.4->2.6)
 *
 *  2009/09/09, Paulo Pedreiras
 * 		ncurses based user-space task scope
 */


#include <unistd.h>				/* For usleep */
#include <ncurses.h>			/* ncurses.h includes stdio.h */
#include <string.h>
#include <signal.h>

#include <rtai_fifos.h>

#include "app_defs.h"		/* Application-wide parameters */


static char quit=0;  /* Flag to control the end of the application */
int ff;              /* FIFO file descriptor */

static unsigned int task_status=0, old_status=0; /* Local image of task status (bit field)*/;

int row,col;			/* store the number of rows and colums of the screen */
long long tact=0;       /* Virtual current time */

void update_task_status(Task_Evt tevt)
{
  if(tevt.task_act=='s')
    task_status |= (0x1 << tevt.task_id);
  else
    task_status &= ~(0x1 << tevt.task_id);
}


void display_update(int c)
{
  unsigned int taskbit = 1; /* Mask for current task */
  int r;
  unsigned char marker=0;			/* Aux vars */

  for(r=0;r<NUM_TASKS;r++) {
    if((old_status & taskbit) == (task_status & taskbit)) {
      if(task_status & taskbit)
	marker=MARK_RUNNING;
      else
	marker=MARK_IDLE;
    }
    else {
      if(task_status & taskbit)
	marker = MARK_GORUNN;
      else
	marker = MARK_GOIDLE;
    }
    mvaddch(FIRST_ROW+STEP_ROW*r,c,marker);
    taskbit=taskbit<<1;
  }

  refresh();

}

void Check_EODisp(int *c)
{
  Task_Evt tevt;
  unsigned int task_trig=TASK_TRIG; 		/* Task for which trigger is done */
  int r;

  if(*c >= col-RIGHT_MARGIN) {

    // Put old trace in gray
    for(r=0;r<NUM_TASKS;r++) {
      mvchgat(FIRST_ROW+STEP_ROW*r,FIRST_COL,-1, A_DIM, 0, NULL);
    }
    *c=FIRST_COL;
    refresh();

    // Wait for a 0->1 transition in channel selected for trigger
     while( task_status & (1<<task_trig) ) {
      read(ff,&tevt,sizeof(tevt));
      update_task_status(tevt);
    }

    while( !(task_status & (1<<task_trig))){
      read(ff,&tevt,sizeof(tevt));
      update_task_status(tevt);
    }

    tact=tevt.evt_time;

  }
}


static void endme (int dummy)
{
  quit=1;
}


// Main program
int main()
{
	// Local vars
	int r,c=FIRST_COL;			/* Row and column to plot graphs */
	unsigned char car;
	unsigned int hold_flag=0;				/* hold flag */

	char mesg_title[]="RTAI task scope, V1.0";		/* message to be appeared on the screen */
	char mesg_keys[]="[h]-hold [r]-resume [q]-quit";

	Task_Evt tevt;

	/* Catch CTRL+C */
	signal(SIGINT, endme);

	// Opens the FIFO
	if( (ff=rtf_open_sized("/dev/rtf0",O_RDONLY,FIFO_SIZE))<0){
	  printf("Error opening /dev/rtf0\n");
	  return -1;
	}

	// NCurses initializations
	initscr();					/* start the curses mode */
	raw();						/* Line buffering disabled	*/
	keypad(stdscr, TRUE);		/* We get F1, F2 etc..		*/
	noecho();					/* Don't echo() while we do getch */
	nodelay(stdscr,TRUE);		/* Non-block getch() */
	getmaxyx(stdscr,row,col);			/* get the number of rows and columns */
	mvprintw(0,(col-strlen(mesg_title))/2,"%s",mesg_title); 		/* print the message at the center of the screen */
	mvprintw(row-1,(col-strlen(mesg_keys))/2,"%s",mesg_keys); 	/* print the message at the bottom of the screen */
	refresh();

	// Processing loop
	attron(A_BOLD);
	c=col;	// Force triggering at first instance
	quit=0; //Quit flag
	while(!quit) {
	  // Handle keyboard
	  car=getch();
	  if(car != ERR) {
	    if(car == 'q' || car == 'Q')
	      quit=1;
	    if(car=='h' || car == 'H')
	      hold_flag=1;
	    if(car=='r' || car == 'r')
	      hold_flag=0;
	  }

	  if(hold_flag)
	    continue;

	  // Updates the display

	  if( read(ff,&tevt,sizeof(tevt)) == sizeof(tevt) ) {
	    if(tact==0)
	      tact=tevt.evt_time;
	    while(tact<tevt.evt_time) {
	      display_update(c++);
	      old_status=task_status;
	      tact+=DISP_STEP_MS*1000*1000;
	      // Check if end of window reached
	      Check_EODisp(&c);
	    }

	    update_task_status(tevt);

	  }

	}

	// Done with ncurses
	endwin();


	/* Close RT FIFO */
	close(ff);

	return 0;
}
